==> exframe-service-1.5.0.tgz: package/lib/memory-watcher.js <==
'use strict';

const os = require('os');
const v8 = require('v8');

const megabytesToBytes = (megabytes) => megabytes * 1024 * 1024;

class MemoryWatcher {
  #logger;
  #running;
  #liveness;
  #readiness;
  #livenessThreshold;
  #readinessThreshold;
  #healthyInterval;
  #unhealthyInterval;
  static #instance;

  /**
   * @param {object} config Configuration options for the new memory watcher instance.
   * @param {number} config.livenessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the liveness probe. Default is 50.
   * @param {number} config.readinessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the readiness probe. Default is 20.
   * @param {number} config.healthyInterval Number of milliseconds between memory checks when all probes
   *  are succeeding. Default is 10000 (10 seconds).
   * @param {number} config.unhealthyInterval Number of milliseconds between memory checks when one or more
   *  probes are failing. Default is 2000 (2 seconds).
   */
  constructor(config = {}) {
    const memoryStatistics = MemoryWatcher.getMemoryStatistics();

    this.#logger = config.logger;
    const { process: { total: systemMemory } } = memoryStatistics;
    this.#livenessThreshold = systemMemory - megabytesToBytes(config.livenessThresholdOffset || 50);
    this.#readinessThreshold = systemMemory - megabytesToBytes(config.readinessThresholdOffset || 20);
    this.#liveness = false;
    this.#readiness = false;
    this.#healthyInterval = config.healthyInterval || 10 * 1000;
    this.#unhealthyInterval = config.unhealthyInterval || 2 * 1000;

    if (this.#logger) {
      this.#logger.info('Service startup memory statistics', { memoryStatistics });
    }
  }

  /**
   * Indicates whether the process should be considered alive and healthy according to memory usage.
   * Intended for use within the kubernetes liveness probe.
   */
  get liveness() {
    return this.#liveness;
  }

  /**
   * Indicates whether the process should be considered ready to receive requests according to memory usage.
   * Intended for use within the kubernetes readiness probe.
   */
  get readiness() {
    return this.#readiness;
  }

  #checkMemory() {
    const memory = MemoryWatcher.getMemoryStatistics();

    if (this.#logger) {
      this.#logger.silly('Current memory usage', { memory });
    }

    if (memory.process.used > this.#livenessThreshold) {
      this.#liveness = false;
      this.#readiness = false;
    } else if (memory.process.used > this.#readinessThreshold) {
      this.#liveness = true;
      this.#readiness = false;
    } else {
      this.#liveness = true;
      this.#readiness = true;
    }

    const nextInterval = this.#liveness && this.#readiness
      ? this.#healthyInterval
      : this.#unhealthyInterval;

    if (this.#running) {
      setTimeout(this.#checkMemory.bind(this), nextInterval);
    }
  }

  /**
   * Starts the process that periodically checks current memory usage at the configured interval to determine health
   * of the process according to configured thresholds.
   */
  start() {
    if (this.#running === true) return;

    if (this.#logger) {
      this.#logger.info('Starting Memory Watcher with the following config', {
        config: {
          livenessThreshold: this.#livenessThreshold,
          readinessThreshold: this.#readinessThreshold,
          healthyInterval: this.#healthyInterval,
          unhealthyInterval: this.#unhealthyInterval
        }
      });
    }

    this.#running = true;
    this.#checkMemory();
  }

  /**
   * Stops the previously started watch process after the next scheduled check.
   */
  stop() {
    this.#running = false;
  }

  /**
   * Gets current memory statistics
   *
   * @typedef {Object} HeapMemoryStats Heap memory statistics
   * @property {number} used Heap memory (in bytes) currently used by the process. Comes from v8.getHeapStatistics().used_heap_size.
   * @property {number} total Heap memory (in bytes) total allocated for the process. Comes from v8.getHeapStatistics().heap_size_limit.
   *
   * @typedef {Object} ProcessMemoryStats
   * @property {number} used Total memory (in bytes) currently used by the process. Comes from process.memoryUsage.rss().
   * @property {number} total Total memory (in bytes) total available for the system. Comes from os.totalmem().
   *
   * @typedef {Object} MemoryStats Current sampled and calculated memory statistics
   * @property {HeapMemoryStats} heap Heap memory statistics
   * @property {ProcessMemoryStats} process Process memory statistics

   * @returns {MemoryStats}
   */
  static getMemoryStatistics() {
    const { heap_size_limit: heapSizeLimit, used_heap_size: heapSizeUsed } = v8.getHeapStatistics();

    return {
      heap: {
        used: heapSizeUsed,
        total: heapSizeLimit
      },
      process: {
        used: process.memoryUsage.rss(),
        total: os.totalmem()
      }
    };
  }

  /**
   * Gets the MemoryWatcher instance or creates one if it doesn't exist
   *
   * @param {object} config Configuration options for the new memory watcher instance.
   * @param {number} config.livenessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the liveness probe. Default is 50.
   * @param {number} config.readinessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the readiness probe. Default is 20.
   * @param {number} config.healthyInterval Number of milliseconds between memory checks when all probes
   *  are succeeding. Default is 10000 (10 seconds).
   * @param {number} config.unhealthyInterval Number of milliseconds between memory checks when one or more
   *  probes are failing. Default is 2000 (2 seconds).
   * @returns
   */
  static getInstance(config) {
    if (!this.#instance) {
      this.#instance = new MemoryWatcher(config);
    }
    return this.#instance;
  }
}

module.exports = MemoryWatcher;

==> exframe-service-1.5.4.tgz: package/lib/memory-watcher.js <==
'use strict';

const os = require('os');
const v8 = require('v8');
const fs = require('fs');

const megabytesToBytes = (megabytes) => megabytes * 1024 * 1024;

class MemoryWatcher {
  #logger;
  #runningTimeout;
  #liveness;
  #readiness;
  #livenessThreshold;
  #readinessThreshold;
  #healthyInterval;
  #unhealthyInterval;
  #totalMemory;
  #lastMemoryLogTimestamp;
  static #instance;

  /**
   * @param {object} config Configuration options for the new memory watcher instance.
   * @param {number} config.livenessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the liveness probe. Default is 50.
   * @param {number} config.readinessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the readiness probe. Default is 20.
   * @param {number} config.healthyInterval Number of milliseconds between memory checks when all probes
   *  are succeeding. Default is 10000 (10 seconds).
   * @param {number} config.unhealthyInterval Number of milliseconds between memory checks when one or more
   *  probes are failing. Default is 2000 (2 seconds).
   */
  constructor(config = {}) {
    this.#logger = config.logger;
    this.#liveness = false;
    this.#readiness = false;
    this.#healthyInterval = config.healthyInterval || 10 * 1000;
    this.#unhealthyInterval = config.unhealthyInterval || 2 * 1000;
    this.#lastMemoryLogTimestamp = 0;

    this.#totalMemory = MemoryWatcher.getTotalMemory();
    this.#livenessThreshold = this.#totalMemory - megabytesToBytes(config.livenessThresholdOffset || 20);
    this.#readinessThreshold = this.#totalMemory - megabytesToBytes(config.readinessThresholdOffset || 50);
  }

  /**
   * Indicates whether the process should be considered alive and healthy according to memory usage.
   * Intended for use within the kubernetes liveness probe.
   */
  get liveness() {
    return this.#liveness;
  }

  /**
   * Indicates whether the process should be considered ready to receive requests according to memory usage.
   * Intended for use within the kubernetes readiness probe.
   */
  get readiness() {
    return this.#readiness;
  }

  #checkMemory() {
    const memory = this.#getMemoryStatistics();

    if (this.#logger) {
      let log = this.#logger.silly;
      if (!this.#lastMemoryLogTimestamp || (Date.now() - this.#lastMemoryLogTimestamp >= 3600000)) {
        log = this.#logger.info;
        this.#lastMemoryLogTimestamp = Date.now();
      }
      log('exframe-service: current memory usage', { memory });
    }

    if (memory.process.used > this.#livenessThreshold) {
      this.#liveness = false;
      this.#readiness = false;
    } else if (memory.process.used > this.#readinessThreshold) {
      this.#liveness = true;
      this.#readiness = false;
    } else {
      this.#liveness = true;
      this.#readiness = true;
    }

    const nextInterval = this.#liveness && this.#readiness
      ? this.#healthyInterval
      : this.#unhealthyInterval;

    this.#runningTimeout = setTimeout(this.#checkMemory.bind(this), nextInterval);
  }

  /**
   * Gets current memory statistics
   *
   * @typedef {Object} HeapMemoryStats Heap memory statistics
   * @property {number} used Heap memory (in bytes) currently used by the process. Comes from v8.getHeapStatistics().used_heap_size.
   * @property {number} total Heap memory (in bytes) total allocated for the process. Comes from v8.getHeapStatistics().heap_size_limit.
   *
   * @typedef {Object} ProcessMemoryStats
   * @property {number} used Total memory (in bytes) currently used by the process. Comes from process.memoryUsage.rss().
   * @property {number} total Total memory (in bytes) total available for the system. Comes from cgroup limits or os.totalmem().
   *
   * @typedef {Object} MemoryStats Current sampled and calculated memory statistics
   * @property {HeapMemoryStats} heap Heap memory statistics
   * @property {ProcessMemoryStats} process Process memory statistics

   * @returns {MemoryStats}
   */
  #getMemoryStatistics() {
    const { heap_size_limit: heapSizeLimit, used_heap_size: heapSizeUsed } = v8.getHeapStatistics();

    return {
      heap: {
        used: heapSizeUsed,
        total: heapSizeLimit
      },
      process: {
        used: process.memoryUsage.rss(),
        total: this.#totalMemory
      }
    };
  }

  /**
   * Starts the process that periodically checks current memory usage at the configured interval to determine health
   * of the process according to configured thresholds.
   */
  start() {
    if (this.#runningTimeout) return;

    if (this.#logger) {
      this.#logger.info('exframe-service: starting memory-watcher with the following config', {
        config: {
          livenessThreshold: this.#livenessThreshold,
          readinessThreshold: this.#readinessThreshold,
          healthyInterval: this.#healthyInterval,
          unhealthyInterval: this.#unhealthyInterval
        }
      });
    }

    this.#checkMemory();
  }

  /**
   * Stops the previously started watch process after the next scheduled check.
   */
  stop() {
    clearTimeout(this.#runningTimeout);
    this.#runningTimeout = undefined;
  }

  /**
   * Get total memory from the current system, with support for limits configured
   * with both cgroup v1 and v2.
   *
   * @returns {function} Total memory for the system
   */
  static getTotalMemory() {
    const cgroupV1 = '/sys/fs/cgroup/memory/memory.limit_in_bytes';
    const cgroupV2 = '/sys/fs/cgroup/memory.max';
    let totalMemory;

    const isUnlimited = (value, unlimitedValue) => value === unlimitedValue;

    if (fs.existsSync(cgroupV1)) {
      const cgroupV1Limit = fs.readFileSync(cgroupV1, 'utf8').trim();
      if (!isUnlimited(cgroupV1Limit, '-1')) {
        totalMemory = cgroupV1Limit;
      }
    } else if (fs.existsSync(cgroupV2)) {
      const cgroupV2Limit = fs.readFileSync(cgroupV2, 'utf8').trim();
      if (!isUnlimited(cgroupV2Limit, 'max')) {
        totalMemory = cgroupV2Limit;
      }
    }

    if (!totalMemory) {
      totalMemory = os.totalmem();
    }

    return totalMemory;
  }

  /**
   * Gets the MemoryWatcher instance or creates one if it doesn't exist
   *
   * @param {object} config Configuration options for the new memory watcher instance.
   * @param {number} config.livenessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the liveness probe. Default is 50.
   * @param {number} config.readinessThresholdOffset Number of megabytes subtracted from total system memory
   *  to calculate the memory threshold above which a service should fail the readiness probe. Default is 20.
   * @param {number} config.healthyInterval Number of milliseconds between memory checks when all probes
   *  are succeeding. Default is 10000 (10 seconds).
   * @param {number} config.unhealthyInterval Number of milliseconds between memory checks when one or more
   *  probes are failing. Default is 2000 (2 seconds).
   * @returns
   */
  static getInstance(config) {
    if (!this.#instance) {
      this.#instance = new MemoryWatcher(config);
    }
    return this.#instance;
  }
}

module.exports = MemoryWatcher;
